{************************************************}
{                                                }
{   To Do List program                           }
{   by Mark Vershel                              }
{   Copyright (c) 1991 by Borland International  }
{                                                }
{************************************************}

program ToDoProgram;

uses Strings, WinTypes, WinProcs, WinDos, WObjects, StdDlgs;

{$R TODO.RES}

const
  cm_New    = 101;     { menu id's }
  cm_Open   = 102;
  cm_Save   = 103;
  cm_SaveAs = 104;
  cm_Exit   = 105;
  cm_Add 		= 201;
  cm_Change	= 202;
  cm_Delete	= 203;
  cm_About  = 901;

  id_A_Button 				= 301;    { dialog ids }
	id_B_Button 				= 302;
	id_C_Button 				= 303;
  id_selectedItem			= 304;
	id_inProcess_Button = 305;
	id_itsDone_Button 	= 306;

  A_priority	  = 1;     { item priorities }
  B_priority	  = 2;
  C_priority	  = 3;
  done_priority = 100;   { If item is "done", it is at bottom of list }

  new_item		= 0;     { states }
  in_process	= 1;
  its_done		= 2;

  itemLen				=	60;			{maximum length of toDo Item }
  WindowTitle		= 'To Do List - ';
  unknownFile		= '<unknown>';
  todoExtension	=	'.tdo';
  genericToDoFile = '*.tdo';

type
  TToDoApplication = object(TApplication)
    procedure InitMainWindow; virtual;
  end;

type
  itemStatusInfo = record    { used for info transfer to/from dialog }
  	itemSelected: array[0..itemLen] of char;
  	priorityA: Bool;
  	priorityB: Bool;
  	priorityC: Bool;
    inProcess: Bool;
    itsDone: Bool;
  end;

type
  PItem = ^TItem;            { stores info on each ToDo item }
  TItem = object(TObject)
    ToDoItem: array[0..itemLen] of char;
    Priority, State: Integer;
    constructor Init(AnItem: Pchar; APriority, AState: Integer);
    constructor Load(var S: TStream);
    procedure Store(var S: TStream);
  end;

type
  PToDoWindow = ^TToDoWindow;
  TToDoWindow = object(TWindow)
    ToDoItems: PCollection;       						{ list of items }
    CurrentItem: PItem;           						{ currently selected item }
    FileName: array[0..fsPathName] of Char;   { name of file }
    ChangesNotSaved, IsNewFile: Boolean;      { add'l info about list }
    yFirstItem: integer;                      { starting point of list }

    constructor Init(AParent: PWindowsObject; ATitle: PChar);
    destructor Done; virtual;
		procedure SetupWindow; virtual;
    function  GetClassName: PChar; virtual;
    procedure GetWindowClass(var AwndClass: TWndClass); virtual;
    function CanClose: Boolean; virtual;

    procedure WMLButtonDown(var Msg: TMessage);
      virtual wm_First + wm_LButtonDown;
    procedure WMLButtonDblClk(var Msg: TMessage);
      virtual wm_First + wm_LButtonDblClk;
    procedure Paint(PaintDC: HDC; var PaintInfo: TPaintStruct); virtual;
    procedure FileNew(var Msg: TMessage);
      virtual cm_First + cm_New;
    procedure FileOpen(var Msg: TMessage);
      virtual cm_First + cm_Open;
    procedure FileSave(var Msg: TMessage);
      virtual cm_First + cm_Save;
    procedure FileSaveAs(var Msg: TMessage);
      virtual cm_First + cm_SaveAs;
    procedure FileExit(var Msg: TMessage);
      virtual cm_First + cm_Exit;
    procedure AddItem(var Msg: TMessage);
      virtual cm_First + cm_Add;
    procedure ChangeItem(var Msg: TMessage);
      virtual cm_First + cm_Change;
    procedure DeleteItem(var Msg: TMessage);
      virtual cm_First + cm_Delete;
    procedure About(var Msg: TMessage);
      virtual cm_First + cm_About;

    { These methods are called from the methods above.  They aren't
      called in response to windows events directly }

    procedure LoadFile;
    procedure SaveFile;
    procedure ReOrderList;
    procedure SelectItem(var Msg: TMessage);
    procedure EditItem;
  end;

type
  PItemStatusDlg = ^TItemStatusDlg;
  TItemStatusDlg = object(TDialog)
  end;

{--------------------------------------------------}
{ Stream construction and registration procedure   }
{--------------------------------------------------}

const
  RItem: TStreamRec = (
    ObjType: 200;
    VmtLink: Ofs(TypeOf(TItem)^);
    Load: @TItem.Load;
    Store: @TItem.Store);

procedure StreamRegistration;
begin
  RegisterType(RCollection);
  RegisterType(RItem);
end;

{--------------------------------------------------}
{ TToDoWindow's method implementations:            }
{--------------------------------------------------}

constructor TToDoWindow.Init(AParent: PWindowsObject; ATitle: PChar);
begin
  TWindow.Init(AParent, ATitle);
  Attr.Menu := LoadMenu(HInstance, 'Menu_1');
  Attr.Style := Attr.Style or ws_VScroll or ws_HScroll;

  Attr.w := 400;   { Why this w x h? Because I like this aspect ratio  }
  Attr.h := 300;

  { Scroller info will be set during Paint }
  Scroller := New(PScroller, Init(@Self, 0, 0, 0, 0));

  ChangesNotSaved := False;
  IsNewFile := True;
  currentItem := nil;

  ToDoItems := New(PCollection, Init(10, 5));
  StreamRegistration;
end;

{--------------------------------------------------}

destructor TToDoWindow.Done;
begin
  Dispose(ToDoItems, Done);     { clean up collection }
  TWindow.Done;
end;

{--------------------------------------------------}

procedure TToDoWindow.SetupWindow;
var
	argCount: integer;
begin
  TWindow.SetupWindow;
  argCount := getArgCount;     { was a file included as a parameter? }
  if argCount > 0 then
  begin
  	getArgStr(FileName, 1, fsPathName);
    loadFile;                  { call loadFile method to load file }
  end;
end;

{--------------------------------------------------}

function TToDoWindow.GetClassName: PChar;
begin
  GetClassName := 'MVToDoWindow';
end;

{--------------------------------------------------}

procedure TToDoWindow.GetWindowClass(var AwndClass: TWndClass);
begin
	TWindow.GetWindowClass(AwndClass);
  AwndClass.hIcon := loadIcon(Hinstance, 'icon_1');
  AwndClass.style := AwndClass.style or cs_dblClks;
end;

{--------------------------------------------------}

function TToDoWindow.CanClose: Boolean;
var
  Reply : Integer;
  Msg: TMessage;
begin
  CanClose := True;
  if ChangesNotSaved then     { be sure to check if something should }
  begin                       { be saved to a file                   }
    MessageBeep(0);

    Reply := MessageBox(HWindow, 'Do you want to save?',
      'To Do List has changed', mb_YesNoCancel or mb_IconQuestion);

    if Reply = id_Yes then
    begin
	    if IsNewFile then FileSaveAs(Msg)
      else SaveFile;
		end
    else if Reply = id_cancel then CanClose := False;
  end;
end;

{--------------------------------------------------}

procedure TToDoWindow.WMLButtonDown(var Msg: TMessage);
begin
  SelectItem(Msg);       { single click selects item }
end;

{--------------------------------------------------}

procedure TToDoWindow.WMLButtonDblClk(var Msg: TMessage);
begin
	SelectItem(Msg);       { double click edits item }
	EditItem;
end;

{--------------------------------------------------}

procedure TToDoWindow.Paint(PaintDC: HDC; var PaintInfo: TPaintStruct);
const
  startX = 10;   { indent 10 pixels }
  heading1 = 'Pri';
  heading2 = 'Items To Do';
  itemStarted = '*';
  itemDone = 'x';
var
  First: boolean;
  s1,s2: array[0..3] of Char;
  textLine: array[0..itemLen] of Char;
  x1, x2, x3, y, ht, wd: Integer;   { These are all set when headings are drawn }
  itemState, itemPriority: Integer;
  tm: TTextMetric;
  oldColor, oldBColor: longInt;

	procedure DrawColumnHeadings;  { internal procedure to draw headings
  																and set first y position }
  begin
    GetTextMetrics(PaintDC, tm);  { determine character height and width }
    ht := tm.tmHeight;
    wd := tm.tmAveCharWidth;

   { set up scroll bars }
    Scroller^.SetUnits(wd, ht);
    Scroller^.SetRange(60, todoItems^.count);   {60 chars is max item length }

   { write out column headings(use lines for underlining) }
    x1 := startX;     { no heading for itemState column }

    x2 := startX + 2*wd;
	 	TextOut(PaintDC, x2, 0, heading1, StrLen(heading1));
    MoveTo(PaintDC, x2, ht);
    LineTo(PaintDC, x2+StrLen(heading1)*wd, ht);
    x2 := x2 + wd;    { indent items in this column by one character}

    x3 := startX + 4*wd + StrLen(heading1)*wd;
    TextOut(PaintDC, x3, 0, heading2, StrLen(heading2));
    MoveTo(PaintDC, x3, ht);
    LineTo(PaintDC, x3+StrLen(heading2)*wd, ht);

    yFirstItem :=(ht*3) div 2;   { determine starting point of items }
    y := yFirstItem;
  end;

  procedure DrawItem(P: PItem); far;    { procedure that draw's each item }
	begin

    { if this is first item, draw column headings }
    if First then
    begin
  		DrawColumnHeadings;
      First := False;
    end;

	 { write out each item as triplet of <status, priority, item> }
	  itemState := P^.state;
	  if itemState = its_done then StrCopy(s1, itemDone)
	  else if itemState = in_process then StrCopy(s1, itemStarted)
		else StrCopy(s1, '');

	  itemPriority := P^.priority;
	  if itemPriority = A_priority then StrCopy(s2, 'A')
		else if itemPriority = B_priority then StrCopy(s2, 'B')
	 	else if itemPriority = C_priority then StrCopy(s2, 'C')
	  else StrCopy(s2, '');   { this occurs normally with "done" item }

	  strCopy(textLine, P^.ToDoItem);

	  if P = currentItem then   { highlight current item in reverse video }
	  begin
      TextOut(PaintDC, x1, y, s1, StrLen(s1));
	    oldBcolor := setBkColor(paintDC, RGB(0,0,0));
	    oldColor := setTextColor(paintDC, oldBcolor);
	   	TextOut(PaintDC, x2, y, s2, StrLen(s2));
      TextOut(PaintDC, x3, y, textLine, StrLen(textLine));
	    setTextColor(PaintDC, oldColor);
  	  setBkColor(PaintDC, oldBColor);
	  end
	  else if itemState= its_done then   { done items are greyed }
	  begin
  	  oldColor := setTextColor(paintDC, RGB(200,200,200));
   		TextOut(PaintDC, x1, y, s1, StrLen(s1));
	   	TextOut(PaintDC, x2, y, s2, StrLen(s2));
  		TextOut(PaintDC, x3, y, textLine, StrLen(textLine));
    	setTextColor(PaintDC, oldColor);
	  end
  	else begin
   		TextOut(PaintDC, x1, y, s1, StrLen(s1));
	   	if itemPriority = A_priority then       { priority A items have A in red }
  	  begin
	  	  oldColor := setTextColor(paintDC, RGB(200,0,0));
    		TextOut(PaintDC, x2, y, s2, StrLen(s2));
	      setTextColor(PaintDC, oldColor);
  	  end
    	else TextOut(PaintDC, x2, y, s2, StrLen(s2));
	  	TextOut(PaintDC, x3, y, textLine, StrLen(textLine));
  	  end;
	  y := y + ht;
	end;

{ this is procedure body for Paint method }

begin
  First := true;      { First item has yet to be drawn }
  ReOrderList;        { sort list by priority order before drawing }
  ToDoItems^.ForEach(@DrawItem);
end;

{--------------------------------------------------}

procedure TToDoWindow.FileNew(var Msg: TMessage);
var
	title: array[0..fsPathName+20] of Char;
begin
  if canClose then
  begin
  	ToDoItems^.FreeAll;
	  currentItem := nil;
		ChangesNotSaved := False;
	  IsNewFile := True;
	  strCopy(title, windowTitle);
    strCat(title, unknownFile);
    setWindowText(Hwindow, title);
  	InvalidateRect(HWindow, nil, True);    { cause redraw }
	end;
end;

{--------------------------------------------------}

procedure TToDoWindow.FileOpen(var Msg: TMessage);
begin
  if CanClose then
    if Application^.ExecDialog(New(PFileDialog,
        Init(@Self, PChar(sd_FileOpen),
        StrCopy(FileName, genericToDoFile)))) = id_Ok then
    begin
      LoadFile;
    end;
end;

{--------------------------------------------------}

procedure TToDoWindow.FileSave(var Msg: TMessage);
begin
  if IsNewFile then FileSaveAs(Msg) else SaveFile;
end;

{--------------------------------------------------}

procedure TToDoWindow.FileSaveAs(var Msg: TMessage);
var
  FileDlg: PFileDialog;
	title: array[0..fsPathName+20] of Char;
begin
  if ToDoItems^.count > 0 then
  begin
	  if IsNewFile then StrCopy(FileName, '');
  	if Application^.ExecDialog(New(PFileDialog,
    	  Init(@Self, PChar(sd_FileSave), FileName))) = id_Ok then
	  	begin
				if (strScan(FileName, '.') = nil)
        then StrCat(FileName, todoExtension);         { attach extension if none given }
	 		  SaveFile;
			  strCopy(title, windowTitle);
    		strCat(title, ansiLower(FileName));
    		setWindowText(Hwindow, title);
      end;
    end
  else MessageBox(Hwindow, 'Nothing to save','Oops',mb_ok or mb_IconExclamation);
end;

{--------------------------------------------------}

procedure TToDoWindow.FileExit(var Msg: TMessage);
var
  FileDlg: PFileDialog;
begin
  if canClose then Done;
end;

{--------------------------------------------------}

procedure TToDoWindow.AddItem(var Msg: TMessage);
var
  InputText: array[0..itemLen] of Char;
  ErrorPos: Integer;
begin
  currentItem := nil;
  editItem;
 	InvalidateRect(HWindow, nil, True);
end;

{--------------------------------------------------}

procedure TToDoWindow.DeleteItem(var Msg: TMessage);
var
  Reply : Integer;
begin
  if currentItem <> nil then
  begin
  	Reply := MessageBox(HWindow, 'Do you want to delete selected item?',
	    'Confirm Delete Operation', mb_YesNoCancel or mb_IconQuestion);
  	if Reply = id_Yes then
	  begin
     	ChangesNotSaved := True;
			ToDoItems^.Delete(CurrentItem);
	    CurrentItem := nil;
	  	InvalidateRect(HWindow, nil, True);
	  end;
  end
  else MessageBox(Hwindow, 'No Item Selected','Oops',mb_ok or mb_IconExclamation);
end;

{--------------------------------------------------}

procedure TToDoWindow.ChangeItem(var Msg: TMessage);
var
  Reply : Integer;
begin
  if currentItem <> nil then editItem
  else MessageBox(Hwindow, 'No Item Selected','Oops',mb_ok or mb_IconExclamation);
end;

{--------------------------------------------------}

procedure TToDoWindow.About(var Msg: TMessage);
begin
  Application^.ExecDialog(New(PDialog, Init(@Self, 'Aboutbox')));
end;

{----------------------------------------------------}
{ Procedures which work on ToDoLists in this window  }
{----------------------------------------------------}

procedure TToDoWindow.LoadFile;
var
  TempColl: PCollection;
  TheFile: TDosStream;
  textLine: array[0..fsPathName+20] of char;
begin
  currentItem := nil;
  TheFile.Init(FileName, stOpen);
  TempColl := PCollection(TheFile.Get);
  TheFile.Done;
  if TempColl <> nil then
  begin
    Dispose(ToDoItems, Done);
    ToDoItems := TempColl;
    ChangesNotSaved := False;
    IsNewFile := False;
	  strCopy(textLine, windowTitle);
    strCat(textLine, ansiLower(FileName));
    setWindowText(Hwindow, textLine);
    InvalidateRect(HWindow, nil, True);
  end
  else begin
  	strCopy(textLine, FileName);
    strCat(textLine, ' does not exist');
  	MessageBox(Hwindow, textLine, 'To Do List: Load Error',mb_ok or mb_IconExclamation);
  	strCopy(FileName, '');
  end;
end;

{--------------------------------------------------}

procedure TToDoWindow.SaveFile;
var
  TheFile: TDosStream;
begin
  TheFile.Init(FileName, stCreate);
  TheFile.Put(ToDoItems);
  TheFile.Done;
  IsNewFile := False;
  ChangesNotSaved := False;
end;

{--------------------------------------------------}

procedure TToDoWindow.ReOrderList;
var
	i, itemCount: integer;
  item1, item2: PItem;
  ChangesMade: boolean;
begin
  ChangesMade := true;
  itemCount := ToDoItems^.count;

	while ChangesMade do       { do simple bubble sort by priority }
	begin
   	ChangesMade := false;
	  i := itemCount-1;

    while (i > 0) do
   	begin
	 	  item2 := ToDoItems^.at(i);
    	item1 := ToDoItems^.at(i-1);

      if item1^.priority > item2^.priority then
	  	begin
    		ToDoItems^.Delete(item1);
       	ToDoItems^.AtInsert(i,item1);
        changesMade := true;
			end;

		  Dec(i);
    end;
  end;
end;

{--------------------------------------------------}

procedure TToDoWindow.selectItem(var Msg: Tmessage);
var
	mouse_y,y,ht: integer;
  tm: TTextMetric;
  theDC: HDC;

 { determine which item is selected; called by ForEach method below }
	procedure isThisTheItem(P: PItem); far;
	begin
	  if y <= mouse_y then
	  	if mouse_y <= y+ht then currentItem := P;
	  y := y + ht;
	end;

begin
  currentItem := nil;         { start with no item selected }
  theDC := getDC(Hwindow);
  getTextMetrics(theDC, tm);
  ht := tm.tmHeight;
  y := yFirstItem;			{ this was set during Paint}
  mouse_y := Msg.LparamHi;        { this is the y coordinate of mouse click }
  mouse_y := mouse_y + Scroller^.yPos*ht;  { adjust for scrolling }
  releaseDC(Hwindow, theDC);

  ToDoItems^.ForEach(@IsThisTheItem);
  InvalidateRect(HWindow, nil, True);   { redraw window to change highlight}
end;

{--------------------------------------------------}

procedure TToDoWindow.editItem;
const
  ItemRecord: ItemStatusInfo = (
    itemSelected: '';
  	priorityA: False;
  	priorityB: False;
  	priorityC: False;
    inProcess: False;
    itsDone: False);
var
	E1: PEdit;
  R1: PRadioButton;
	C1: PCheckBox;
	Thedlg: PitemStatusDlg;
  Reply: integer;

begin
 { set up transfer record, display dialog window, allow user to
   edit in that window, and change item if necessary
 }
 	strCopy(itemRecord.itemSelected, '');   { blank out transfer record }
  ItemRecord.priorityA := false;
  ItemRecord.priorityB := false;
  ItemRecord.priorityC := false;
  ItemRecord.inProcess := false;
	ItemRecord.itsDone   := false;

  if currentItem <> nil then  { if item selected, use its info }
  begin
		strCopy(itemRecord.itemSelected, currentItem^.ToDoItem);

    case CurrentItem^.Priority of
    	A_priority: ItemRecord.priorityA := true;
    	B_priority: ItemRecord.priorityB := true;
    	C_priority: ItemRecord.priorityC := true;
    end;

    if CurrentItem^.State = in_process
    then ItemRecord.inProcess := true;

    if CurrentItem^.State = its_done
	  then ItemRecord.itsDone := true;
  end
  else itemRecord.priorityC := True;   { else use blank item with C priority }

  { display dialog }
	TheDlg := New(PItemSTatusDlg, Init(@self, 'ITEM_STATUS'));
  New(E1, InitResource(TheDlg, id_SelectedItem, SizeOf(ItemRecord.itemSelected)));
  New(R1, InitResource(TheDlg, id_A_Button));
	New(R1, InitResource(TheDlg, id_B_Button));
	New(R1, InitResource(TheDlg, id_C_Button));
	New(C1, InitResource(TheDlg, id_inProcess_Button));
	New(C1, InitResource(TheDlg, id_itsDone_Button));
	TheDlg^.TransferBuffer := @ItemRecord;

	Reply := Application^.ExecDialog(TheDlg);

  if Reply = id_ok then    { user hit OK button }
  begin
    if itemRecord.ItemSelected[0] = #0      { user must enter an item }
    then MessageBox(HWindow, 'To Do Item entered is blank',
      							'Oops', mb_Ok or mb_IconExclamation)
    else begin
      changesNotSaved := True;
  	  if CurrentItem = nil then   { create new item; i.e., not an edit }
 	   	begin
				currentItem := New(PItem, Init(ItemRecord.ItemSelected, C_priority, new_item));
      	ToDoItems^.Insert(currentItem);
    	end;

     { copy info from dialog }

	  	strCopy(CurrentItem^.ToDoItem, ItemRecord.ItemSelected);

			if ItemRecord.priorityA
			then CurrentItem^.Priority := A_Priority
			else if ItemRecord.priorityB
	 		then CurrentItem^.Priority := B_Priority
	 		else CurrentItem^.Priority := C_Priority;

	 	 	if ItemRecord.itsDone then
	 	 	begin
	   		CurrentItem^.State := its_done;
  	  	CurrentItem^.Priority := done_priority;
		  end
			else if ItemRecord.inProcess
		 	then CurrentItem^.State := in_Process
	    else CurrentItem^.State := new_item;

	  	invalidateRect(HWindow, nil, True);    { cause a redraw }
    end;
  end;
end;

{--------------------------------------------------}
{ TItem's method implementations:                  }
{--------------------------------------------------}

constructor TItem.Init(AnItem: PChar;	APriority, AState: Integer);
begin
  strCopy(ToDoItem, AnItem);
  Priority := APriority;
  State 	 := AState;
end;

constructor TItem.Load(var S: TStream);
begin
  S.Read(ToDoItem, SizeOf(ToDoItem));
  S.Read(Priority, SizeOf(Priority));
  S.Read(State, SizeOf(State));
end;

procedure TItem.Store(var S: TStream);
begin
  S.Write(ToDoItem, SizeOf(ToDoItem));
  S.Write(Priority, SizeOf(Priority));
  S.Write(State, SizeOf(State));
end;

{--------------------------------------------------}
{ TToDoApplication's method implementations:       }
{--------------------------------------------------}

procedure TToDoApplication.InitMainWindow;
var
	title: array[0..fsPathName+20] of char;
begin
 { To Do List is unknown at initialization}
	strCopy(title, windowTitle);
  strCat(title, unknownFile);
  MainWindow := New(PToDoWindow, Init(nil, title));
end;


{--------------------------------------------------}
{ Main program:                                    }
{--------------------------------------------------}

var
  ToDoApp : TToDoApplication;

begin
  ToDoApp.Init('ToDoProgram');
  ToDoApp.Run;
  ToDoApp.Done;
end
.
